-- Part of monadic prompt lazy assertions
-- 
-- Functions for observing when data (including function values) becomes 
-- evaluated.
-- For each observed data structure an eval tree is created and updated.
-- Whenever parts of the data structure are evaluated the eval tree grows and
-- actions stored in the eval tree are executed (they are assumed to terminate).

module Observer 
  (Observe(obs),observe,oPrim,o0,o1,o2,o3
  ,EvalTree(..),EvalTreeRef,emptyUnevalRef,showWFail,coerce) 
  where

import Data.IORef
import System.IO.Unsafe
import List
import PosTrie


-- -------------------------------------------------------------------------------
-- Show an eval tree as its term.

markB :: String -> String
markB str = "\ESC[7m" ++ str ++ "\ESC[0m"


-- Show eval tree highlighting given positions.
showWFail :: PosSet -> EvalTree -> IO String
showWFail ps (Eval p fName []) = 
  return $ if elemPosSet p ps then markB fName else fName
showWFail ps (Eval p fName argRefs) = do
  args <- mapM readIORef argRefs
  shownArgs <- mapM (showWFail ps) args
  return $ 
    let mFun = if elemPosSet p ps then markB else id in
      case fName of
        ":"   -> head shownArgs ++ mFun ":" ++ (shownArgs!!1)
        "(,)" -> mFun "(" ++ head shownArgs ++ mFun "," ++
                 (shownArgs!!1) ++ mFun ")"
        "->"  -> head shownArgs ++ mFun " --> " ++ (shownArgs!!1)
        _     -> "("++concat (intersperse " "
                      (mFun fName : shownArgs))++")"
showWFail ps e@(Fun _ _ _ _ _ _) = do
  list <- funList e
  return ("(" ++ concat (intersperse " | " list) ++ ")")
  where
  funList :: EvalTree -> IO [String]
  funList (Fun p ra rr _ _ rfuns) = do
    a <- readIORef ra
    r <- readIORef rr
    funs <- readIORef rfuns
    sa <- showWFail ps a
    sr <- showWFail ps r
    sfuns <- funList funs
    let mFun = if elemPosSet p ps then markB else id 
    return ((sa ++ mFun " -> " ++ sr) : sfuns)
  funList _ = return []
showWFail _ (Uneval _) = return "_"

-- -------------------------------------------------------------------------------
-- Define the eval tree

data EvalTree = Eval Pos String [EvalTreeRef]
              | Fun Pos EvalTreeRef EvalTreeRef () () EvalTreeRef
              | Uneval (IORef (IO ()))

type EvalTreeRef = IORef EvalTree

type Obs a = a -> EvalTreeRef -> Pos -> a

class Observe a where
  obs :: Obs a

instance Show (a -> b) where
  show _ = "FUN"

oPrim :: Show a => Obs a
oPrim x r p = unsafePerformIO $ do
      --seq x (return ())
      if show x == show x then (return ()) else (return ())
      mkEvalTreeCons (show x) r 0 p
      return x

o0 :: String -> Obs a
o0 str cons r p = unsafePerformIO $ do
      mkEvalTreeCons str r 0 p
      return cons

o1 :: Observe a => String -> (a -> b) -> a -> EvalTreeRef -> Pos -> b
o1 str cons vA r p = unsafePerformIO $ do
      [aRef] <- mkEvalTreeCons str r 1 p
      return (cons (obs vA aRef (1/:p)))


o2 :: (Observe a,Observe b) =>
      String -> (a -> b -> c) -> a -> b -> EvalTreeRef -> Pos -> c
o2 str cons vA vB r p = unsafePerformIO $ do
      [aRef,bRef] <- mkEvalTreeCons str r 2 p
      return (cons (obs vA aRef (1/:p))
                   (obs vB bRef (2/:p))) 

o3 :: (Observe a,Observe b,Observe c) => String -> (a -> b -> c -> d)
      -> a -> b -> c -> EvalTreeRef -> Pos -> d
o3 str cons vA vB vC r p = unsafePerformIO $ do
      [aRef,bRef,cRef] <- mkEvalTreeCons str r 3 p
      return (cons (obs vA aRef (1/:p))
                   (obs vB bRef (2/:p))
                   (obs vC cRef (3/:p)))

emptyUnevalRef :: IO EvalTreeRef
emptyUnevalRef = newIORef (return ()) >>= return . Uneval >>= newIORef

mkEvalTreeCons :: String -> EvalTreeRef -> Int -> Pos -> IO [EvalTreeRef]
mkEvalTreeCons str r n p = do
  refs <- mapM (const emptyUnevalRef) [1..n]
  Uneval ref <- readIORef r
  act <- readIORef ref
  writeIORef r (Eval p str refs)
  act
  return refs

instance (Observe a,Observe b) => Observe (a,b) where
  obs (a,b) = o2 "(,)" (,) a b

instance (Observe a,Observe b,Observe c) => Observe (a,b,c) where
  obs (a,b,c) = o3 "(,,)" (,,) a b c

instance Observe a => Observe [a] where
  obs (x:xs) = o2 ":" (:) x xs
  obs []     = o0 "[]" []

instance Observe Int where
  obs = oPrim

instance Observe Char where
  obs = oPrim

instance (Observe a,Observe b) => Observe (a -> b) where
  obs f r p vA = unsafePerformIO $ do
    evalTreeNode <- readIORef r
    case evalTreeNode of
      Fun _ _ _ _ _ nextFunRef -> do
        nextFun <- readIORef nextFunRef
        return (obs f nextFunRef (2/:p) vA)
      Uneval actRef -> do
        aRef <- emptyUnevalRef
        bRef <- emptyUnevalRef
	funRef <- emptyUnevalRef
        let vB = f (obs vA aRef (1/:1/:p))
        act <- readIORef actRef
        writeIORef r (Fun p aRef bRef (coerce vA) (coerce vB) funRef)
        act 
        return (obs vB bRef (2/:1/:p))

observe :: Observe a => a -> IO (EvalTreeRef,a)
observe x = do
  r <- emptyUnevalRef
  return (r,obs x r emptyPos)

-- -------------------------------------------------------
-- Nasty hack to change types arbitrarily
-- Used to store arguments and results of functions in the eval tree.

coerce :: a -> b
coerce x = unsafePerformIO (writeIORef r x >> readIORef r)
  where
  r = unsafePerformIO (newIORef undefined)
