-- Part of monadic prompt lazy assertions
--
-- Data type for positions in a tree structure and
-- implementation of a set of positions.

module PosTrie 
  (Pos,(/:),emptyPos
  ,PosSet,emptyPosSet,elemPosSet,insertPosSet,oneElemPosSet,unionPosSets
  ,posListToPosSet) 
  where

infixr 5 /:

data Pos = Pos [Int]
         | NoPos
  deriving (Show,Eq)

(/:) :: Int -> Pos -> Pos
i /: (Pos xs) = Pos (i:xs)
_ /: NoPos    = NoPos

emptyPos :: Pos
emptyPos = Pos []

data PosSet = PosSet Bool (IntTrie PosSet)
  -- Boolean states whether empty position is an element.
  -- IntTrie is actually list of subtries for some position digits.

type IntTrie a = [(Int,a)]

emptyPosSet :: PosSet
emptyPosSet = PosSet False []

elemPosSet :: Pos -> PosSet -> Bool
elemPosSet NoPos s = False
elemPosSet (Pos xs) s = elemPosSet' (reverse xs) s
  where elemPosSet' :: [Int] -> PosSet -> Bool
        elemPosSet' [] (PosSet b _) = b
	elemPosSet' (x:xs) (PosSet _ succs) =
          maybe False  (elemPosSet' xs) (lookup x succs)

insertPosSet :: Pos -> PosSet -> PosSet
insertPosSet NoPos s = s
insertPosSet (Pos xs) s = insertPosSet' (reverse xs) s
  where insertPosSet' :: [Int] -> PosSet -> PosSet
        insertPosSet' [] (PosSet _ succs) = PosSet True succs
	insertPosSet' (x:xs) (PosSet b succs) =
          PosSet b (modifyPosSet x (insertPosSet' xs) succs)

-- internal helper function
modifyPosSet :: Int -> (PosSet -> PosSet) -> IntTrie PosSet -> IntTrie PosSet
modifyPosSet n f [] = [(n,f emptyPosSet)]
modifyPosSet n f ((m,x):xs) = 
  if n<m then (n,f emptyPosSet):(m,x):xs
         else if n==m then (n,f x):xs
		      else (m,x):modifyPosSet n f xs

oneElemPosSet :: Pos -> PosSet
oneElemPosSet p = insertPosSet p emptyPosSet

unionPosSets :: PosSet -> PosSet -> PosSet
unionPosSets (PosSet b1 succs1) (PosSet b2 succs2) =
  PosSet (b1||b2) (merge succs1 succs2)
  where merge :: IntTrie PosSet -> IntTrie PosSet -> IntTrie PosSet
	merge [] ys = ys
	merge xs [] = xs
	merge ((kx,x):xs) ((ky,y):ys) =
          if kx<ky
            then (kx,x):merge xs ((ky,y):ys)
	    else if kx==ky
                   then (kx,unionPosSets x y):merge xs ys
                   else (ky,y):merge ((kx,x):xs) ys

posListToPosSet :: [Pos] -> PosSet
posListToPosSet = foldr insertPosSet emptyPosSet